import json
from unittest import TestCase

import respx
from httpx import URL, Response
from tuprolog.theory.parsing import parse_theory

from lpaas_client import TheoryName
from lpaas_client.core.exception.exception import Unauthorized
from lpaas_client.sync.routes.auth import AuthClient
from lpaas_client.core.model.auth import AuthData
from lpaas_client.sync.routes.theories import TheoriesClient

base_url = URL('http://localhost:8080')


class TestAuthClient(TestCase):
    client: AuthClient

    @classmethod
    def setUp(cls) -> None:
        cls.client = AuthClient(base_url)

    def test_can_create_client(self):
        self.assertIsInstance(self.client, AuthClient)

    @respx.mock
    def test_authenticate_with_context(self):
        request = respx.post(base_url.join('/lpaas/auth')).mock(
                return_value=Response(status_code=200, json={'token': 'jwt token value'}))

        with self.client:
            token = self.client.authenticate(AuthData('User', 'pass'))

        # token = self.client.authenticate(AuthData('User', 'pass'))
        self.assertTrue(request.called)
        self.assertEqual(json.loads(request.calls.last.request.content.decode('utf-8')),
                         {'username': 'User', 'password': 'pass'})
        self.assertEqual('jwt token value', token)

    @respx.mock
    def test_authenticate(self):
        request = respx.post(base_url.join('/lpaas/auth')).mock(
                return_value=Response(status_code=200, json={'token': 'jwt token value'}))
        token = self.client.authenticate(AuthData('User', 'pass'))
        self.assertTrue(request.called)
        self.assertEqual(json.loads(request.calls.last.request.content.decode('utf-8')),
                         {'username': 'User', 'password': 'pass'})
        self.assertEqual('jwt token value', token)

    @respx.mock
    def test_get_token(self):
        token = 'jwt token value'
        respx.post(base_url.join('/lpaas/auth')).mock(
                return_value=Response(status_code=200, json={'token': token}))
        self.client.authenticate(AuthData('User', 'pass'))
        self.assertEqual(self.client.get_token(), token)

    @respx.mock
    def test_get_or_create_token(self):
        token = 'jwt token value'
        respx.post(base_url.join('/lpaas/auth')).mock(
                return_value=Response(status_code=200, json={'token': token}))
        self.assertEqual(self.client.get_or_create_token(AuthData('User', 'pass')), token)

    @respx.mock
    def test_add_jwt_token(self):
        token = 'jwt token value'
        respx.post(base_url.join('/lpaas/auth')).mock(
                return_value=Response(status_code=200, json={'token': token}))
        self.client.authenticate(AuthData('User', 'pass'))
        self.assertEqual(self.client._add_jwt_token(), {'Authorization': token})

    @respx.mock
    def test_raise_exception_with_no_user_data(self):
        with self.assertRaises(ValueError):
            self.client.get_or_create_token()

        with self.assertRaises(ValueError):
            self.client.authenticate()

        with self.assertRaises(ValueError):
            self.client.get_token()

    @respx.mock
    def test_raise_exception_with_failed_auth(self):
        respx.post(base_url.join('/lpaas/auth')) % 401

        with self.assertRaises(Unauthorized):
            self.client.authenticate(AuthData('User', 'Pass'))


class TestAuthAutomaticAuth(TestCase):

    @classmethod
    @respx.mock
    def setUp(cls) -> None:
        cls.client = TheoriesClient(base_url=base_url)

        # authentication
        respx.post(base_url.join('/lpaas/auth')).mock(
                return_value=Response(status_code=200, json={'token': 'jwt_token'}))

        cls.client.authenticate(AuthData('user', 'password'))

    @respx.mock
    def test_create_theory_with_request_reject(self):
        theory_name = TheoryName('theory_name')
        theory_str = 'a.\np(a).\np(a, b).\n'
        theory = parse_theory(theory_str)

        request = respx.post(base_url.join('/lpaas/theories')) % 401
        auth_request = respx.post(base_url.join('/lpaas/auth')).mock(
                return_value=Response(status_code=200, json={'token': 'jwt_token'}))

        with self.assertRaises(Unauthorized):
            self.client.add_new_theory(theory_name, theory)
        self.assertTrue(auth_request.called)
        self.assertTrue(request.called)
        self.assertEqual(len(request.calls), 2)
